#!/usr/bin/env python
# -*- coding: utf_8 -*
# (C) Copyright 2012 Nuxeo SAS <http://nuxeo.com>
# Authors: Benoit Delbosc <ben@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
from datetime import timedelta
from util import time_to_datetime
from util import duration_to_second
from util import str2id


class Build(object):
    """ Container for activity information"""
    def __init__(self, url, host, name, build_number, start, duration, status, downstream,
                 base_url, trigger):
        self.url = url
        self.host = host
        self.name = name
        self.build_number = build_number
        self.start = start
        self.duration = duration
        self.status = status
        self.downstream = downstream
        self.children = []
        self.base_url = base_url
        self.trigger = trigger
        self.start_t = time_to_datetime(start)
        self.duration_s = duration_to_second(duration)
        self.stop_t = self.start_t + timedelta(seconds=self.duration_s)

    def getId(self):
        return str2id("%s %s" % (self.name, self.build_number))

    def color(self):
        if self.status == 'Success':
            return "blue"
        if self.status == 'Failure':
            return "red"
        if self.status == 'Unstable':
            return "gold"
        return "black"

    def full_url(self):
        return self.base_url + self.url

    def __repr__(self):
        return 'URL: "%s"\n\tname: %s\n\tbuild #: %s\n\thost: %s\n\tstart: %s\n\tstop: %s\n\tduration: %s\n\tstatus: %s\n\tdownstream build: %d\n' % (
            self.url, self.name, self.build_number, self.host, self.start, self.stop_t, self.duration, self.status,
            len(self.downstream))



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.model
   ~~~~~~~~~~~~~~

   This module contains the data models for the Jenkviz.
"""
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
from datetime import timedelta
from sqlalchemy import Column, String, Integer, DateTime
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from util import time_to_datetime
from util import duration_to_second
from util import str2id

Base = declarative_base()


def init_db(engine):
    Base.metadata.create_all(engine)


def open_db(options):
    engine = create_engine('sqlite:///' + options.database, echo=options.verbose)
    Session = sessionmaker(engine)
    db = Session()
    init_db(engine)
    return db


def close_db(db):
    if db:
        db.close()


def get_build(db, url):
    build = db.query(Build).filter_by(url=url).first()
    if build:
        build.children = []
    return build


def list_builds(db):
    return db.query(Build).all()


def save_build(db, build):
    db.add(build)
    db.commit()


def update_build(db, build):
    db.merge(build)
    db.commit()


class Build(Base):
    __tablename__ = "build"

    url = Column(String(256), primary_key=True)
    base_url = Column(String(128))
    host = Column(String(64))
    name = Column(String(128))
    build_number = Column(Integer)
    start = Column(String(32))
    duration = Column(String(32))
    status = Column(String(32))

    start_t = Column(DateTime)
    stop_t = Column(DateTime)
    duration_s = Column(Integer)
    trigger = Column(String(64))
    downstream = Column(String(4096))
    upstream = Column(String(512))

    def __init__(self, url, host, name, build_number, start, duration, status, downstream,
                 base_url, trigger, upstream):
        self.url = url
        self.host = host
        self.name = name
        self.build_number = build_number
        self.start = start
        self.duration = duration
        self.status = status
        self.children = []
        self.base_url = base_url
        self.trigger = trigger
        self.start_t = time_to_datetime(start)
        self.duration_s = duration_to_second(duration)
        if self.start_t and self.duration_s:
            self.stop_t = self.start_t + timedelta(seconds=self.duration_s)
        else:
            self.stop_t = None
        self.downstream = ','.join(downstream)
        self.upstream = ','.join(upstream)

    def __repr__(self):
        return '''URL: "%s"\n\tname: %s\n\tbuild #: %s\n\thost: %s\n\tstart: %s\n\tstop: %s
\tduration: %s\n\tstatus: %s\n\tdownstream build: %d\n\tupstream: %s\n''' % (
            self.url, self.name, self.build_number, self.host, self.start, self.stop_t, self.duration, self.status,
            len(self.get_downstream()), self.upstream)

    def getId(self):
        return str2id("%s %s" % (self.name, self.build_number))

    def color(self):
        if self.status == 'Success':
            return "blue"
        if self.status == 'Failed':
            return "red"
        if self.status == 'Unstable':
            return "gold"
        if self.status == 'Unknown':
            return "brown"
        if self.status == 'Aborted':
            return "dimgray"
        return "black"

    def full_url(self):
        return self.base_url + self.url

    def get_downstream(self):
        if self.downstream:
            return self.downstream.split(',')
        return []

    def get_upstream(self):
        if self.upstream:
            return self.upstream.split(',')
        return []



#!/usr/bin/env python
# -*- coding: utf_8 -*
"""
Try to do something usefull with a jmeter output.
"""
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
import os
import sys
from optparse import OptionParser, TitledHelpFormatter
from util import get_version, init_logging
from command import *

DEFAULT_DB = "~/jenkviz.db"
DEFAULT_LOG = "~/jenkviz.log"

USAGE = """jenkviz [--version] [--logfile=LOGFILE] [--database=DATABASE] COMMAND [OPTIONS] [ARGUMENT]

COMMANDS:

  crawl [OPTIONS] URL
     crawl the jenkins build produces a SVG file.

"""


def main(argv=sys.argv):
    """Main test"""
    global USAGE
    parser = OptionParser(USAGE, formatter=TitledHelpFormatter(),
                          version="jenkviz %s" % get_version())
    parser.add_option("-v", "--verbose", action="store_true",
                      help="Verbose output")
    parser.add_option("-l", "--logfile", type="string",
                      default=os.path.expanduser(DEFAULT_LOG),
                      help="Log file path")
    parser.add_option("-d", "--database", type="string",
                      default=os.path.expanduser(DEFAULT_DB),
                      help="SQLite db path")
    parser.add_option("-o", "--output", type="string",
                      help="SVG output file")
    parser.add_option("--from-file", type="string",
                      help="Use html files in the the FROM_FILE directory instead of querying jenkins server.")
    parser.add_option("--to-file", type="string",
                      help="Save jenkins page into the TO_FILE directory.")
    parser.add_option("-r", "--reverse", action="store_true",
                      default=False,
                      help="Reverse crawl")
    parser.add_option("--direct", action="store_true",
                      default=False,
                      help="Display only direct upstream dependencies")
    parser.add_option("--explore", action="store_true",
                      default=False,
                      help="Display downstream build with external upstream")
    parser.add_option("-u", "--update", action="store_true",
                      default=False,
                      help="Always fetch build from server (update local database)")

    options, args = parser.parse_args(argv)
    if options.explore:
        options.direct = False
    init_logging(options)
    if len(args) == 1:
        parser.error("Missing command")
    cmd = args[1]
    fn = globals()['cmd_' + cmd]
    ret = fn(args[2:], options)
    return ret


if __name__ == '__main__':
    ret = main()
    sys.exit(ret)



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.command
   ~~~~~~~~~~~~~~~~

   Crawl a Jenkins to extract builds flow.
"""
import os
import logging
from model import open_db, close_db, list_builds
from crawl import Crawl
from graphviz import graphviz


def cmd_list(args, options):
    db = open_db(options)
    list_builds(db)
    close_db(db)
    return 0


def cmd_crawl(args, options):
    if len(args) != 1:
        logging.error("Missing build URL")
        return 1
    if options.to_file and not os.path.exists(options.to_file):
        os.mkdir(options.to_file)
    if options.from_file and not os.path.exists(options.from_file):
        os.mkdir(options.from_file)
    db = open_db(options)
    crawl = Crawl(db, options)
    if options.reverse:
        roots = crawl.reverse_crawl(args[0])
    else:
        roots = crawl.crawl(args[0])
    close_db(db)
    stat = roots[0].extra
    logging.info("Started: %s\n\tend: %s\n\telapsed: %s\n\tduration: %ss\n\tNb builds: %s\n\ttrhoughput: %s\n" % (
            stat['start'], stat['stop'], stat['elapsed'], stat['duration'], stat['count'], stat['throughput']))
    if not options.output:
        svg_file = roots[0].getId() + ".svg"
    else:
        svg_file = options.output
    graphviz(roots, svg_file)
    logging.info("%s generated." % svg_file)
    return 0


def cmd_info(args, options):
    if len(args) != 1:
        logging.error('Missing bid')
        return 1
    db = open_db(options)
    # bencher = Bencher.getBencherForBid(db, options, args[0])
    # print """bid: %(bid)s, from %(start)s to %(end)s, samples: %(count)d, errors: %(error)d""" % bencher.getInfo(args[0])
    close_db(db)
    return 0


def cmd_report(args, options):
    if len(args) != 1:
        logging.error('Missing bid')
        return 1
    if not options.output:
        logging.error('Missing --output option')
        return 1
    db = open_db(options)
    # report = Report(db, options)
    # report.buildReport(args[0])
    close_db(db)
    return 0



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.graphviz
   ~~~~~~~~~~~~~~~~

   Generate a graphviz dot file and svg files from a build tree.
"""
from commands import getstatusoutput
from datetime import timedelta


def graphviz(roots, svg_file):
    """Create a fpath svg from build tree."""
    root = roots[0]
    dot_file = svg_file.replace('.svg', '.dot')
    out = open(dot_file, "w+")
    out.write("""digraph g {
graph [rankdir=LR];
node [fontsize="16" shape="record"];
info [label="start: %s|stop: %s|elapsed: %s|duration: %s|number of builds: %s|throughput: %.2f%%"];
""" % (root.extra['start'], root.extra['stop'], root.extra['elapsed'], timedelta(seconds=root.extra['duration']),
       root.extra['count'], root.extra['throughput']))

    visited = []
    for root in roots:
        _graphviz_recurse(root, out, visited)
    out.write("}\n")
    out.close()
    _make_svg(dot_file, svg_file)


def _graphviz_recurse(parent, out, visited):
    if parent.url in visited:
        return
    visited.append(parent.url)
    out.write('%s [label="%s #%s|%s %s|%s" color=%s URL="%s"]\n' % (
        parent.getId(), parent.name, parent.build_number, str(parent.start_t)[11:], parent.host, parent.duration,
        parent.color(), parent.full_url()))

    if parent.trigger:
        out.write('%s [color=orange style=filled]\n%s -> %s\n' % (
                parent.trigger, parent.trigger, parent.getId()))
    if not parent.children:
        return
    out.write("%s -> {" % parent.getId())
    for build in parent.children:
        if parent.url in build.get_upstream():
            out.write(build.getId() + ";")
    out.write("}\n")
    for build in parent.children:
        if parent.url not in build.get_upstream():
            out.write('%s -> %s [color=orange]\n' % (parent.getId(), build.getId()))
    for build in parent.children:
        _graphviz_recurse(build, out, visited)


def _make_svg(dot_file, svg_file):
    cmd = "dot -Tsvg %s -o %s" % (dot_file, svg_file)
    ret, output = getstatusoutput(cmd)
    if ret != 0:
        raise RuntimeError("Failed to run dot cmd: " + cmd +
                           "\n" + str(output))



#!/usr/bin/env python
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com/>"
__version__ = '0.2.0'
"""jenkviz package setup"""
from setuptools import setup, find_packages


setup(
    name="jenkviz",
    version=__version__,
    description="Crawl a jenkins build and report stats and graphs about the build flow.",
    long_description=''.join(open('README.txt').readlines()),
    author="Benoit Delbosc",
    author_email="bdelbosc@nuxeo.com",
    url="http://pypi.python.org/pypi/jenkviz",
    download_url="http://pypi.python.org/packages/source/t/jenkviz/jenkviz-%s.tar.gz" % __version__,
    packages=find_packages(),
    license='GPL',
    keywords='jenkins hudson CI chart',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Quality Assurance',
    ],
    # setuptools specific keywords
    install_requires=['requests', 'sqlalchemy'],
    zip_safe=True,
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'jenkviz = jenkviz.main:main'],
    },
)



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.crawl
   ~~~~~~~~~~~~~~~~

   Crawl a Jenkins to extract builds flow.
"""
import os
import re
import logging
from urlparse import urlparse
from requests import get as http_get
from model import Build
from util import str2id
from util import extract_token
from model import get_build
from model import save_build
from model import update_build


class Crawl(object):
    def __init__(self, db, options):
        self.db = db
        self.options = options
        self.builds = {}
        self.count = 0
        self.root = None

    def crawl(self, url):
        self.url = url
        u = urlparse(url)
        self.path_root = u.path
        self.server_url = url[:-len(self.path_root)]
        self._crawl(self.path_root)
        self.clean(self.root)
        self.root.extra = self.stats()
        return [self.root]

    def _crawl(self, url):
        parent = self.get_build(url)
        if self.root is None:
            self.root = parent
        print parent
        if not len(parent.get_downstream()):
            return
        for child_url in parent.get_downstream():
            known = False
            if child_url in self.builds.keys():
                known = True
            child = self.get_build(child_url)
            if not self.options.direct or parent.url in child.get_upstream():
                parent.children.append(child)
            if not known:
                self._crawl(child_url)

    def reverse_crawl(self, url):
        self.url = url
        u = urlparse(url)
        self.path_leaf = u.path
        self.server_url = url[:-len(self.path_leaf)]
        self.roots = []
        self._reverse_crawl(self.path_leaf)
        # TODO: disable cache and craw from roots
        # for root in self.roots:
        #    self._crawl(root.url)
        self.roots[0].extra = self.stats()
        return self.roots

    def _reverse_crawl(self, url):
        child = self.get_build(url)
        print child
        if len(child.get_upstream()) == 0:
            self.roots.append(child)
            return
        for parent_url in child.get_upstream():
            known = False
            if parent_url in self.builds.keys():
                known = True
            parent = self.get_build(parent_url)
            parent.children.append(child)
            if not known:
                self._reverse_crawl(parent_url)

    def get_build(self, url):
        self.count += 1
        # 1. fetch from cache
        if url in self.builds.keys():
            return self.builds[url]
        ret = None
        # 2. fetch from local db
        ret = self.fetch_build_from_db(url)
        # 3. fetch jenkins page
        if ret is None or self.options.update:
            build = self.fetch_build(url)
            if build is None:
                return None
            # 3.1 persist build
            if ret is None:
                self.save_build_to_db(build)
            else:
                self.update_build_to_db(build)
            ret = build
        # 4. update cache
        self.builds[url] = ret
        return ret

    def fetch_build(self, url):
        if self.options.from_file:
            body = self.fetch_build_from_file(url)
        else:
            body = self.fetch_build_from_server(url)
        ret = self.parse_build(url, body)
        if ret and self.options.to_file:
            self.save_build_to_file(ret, body)
        return ret

    def fetch_build_from_db(self, url):
        return get_build(self.db, url)

    def save_build_to_db(self, build):
        save_build(self.db, build)

    def update_build_to_db(self, build):
        update_build(self.db, build)

    def fetch_build_from_server(self, url):
        response = http_get(self.server_url + url)
        if response.status_code != 200:
            logging.error('Failure: %s%s return %s' % (self.server_url, url, response.status_code))
            return "ERROR: %s" % response.status_code
        return response.text

    def fetch_build_from_file(self, url):
        dir_path = self.options.from_file
        build_id = str2id(' '.join([i for i in url.split('/') if i][-2:]))
        file_path = os.path.join(dir_path, build_id + '.txt')
        body = '\n'.join(open(file_path).readlines())
        return body

    def save_build_to_file(self, build, body):
        dir_path = self.options.to_file
        build_id = build.getId()
        file_path = os.path.join(dir_path, build_id + '.txt')
        open(file_path, 'w+').write(body.encode('utf-8'))

    def parse_build(self, url, body):
        if body.startswith('ERROR'):
            name = url.split('/')[-3]
            build_number = url.split('/')[-2]
            return Build(url, body, name, build_number, None, None, 'Unknown', [],
                         self.server_url, '', [])
        name = extract_token(body, '<title>', ' ')
        h1 = extract_token(body, '<h1>', '</h1>')
        status = extract_token(h1, 'alt="', '"')
        build_number = extract_token(h1, 'Build #', '\n')
        start = extract_token(h1, "(", ")")
        duration = extract_token(body, '/buildTimeTrend">', '</a')
        host = extract_token(body, '<a href="/jenkins/computer/', '"')
        downstream_builds = extract_token(body, 'h2>Downstream Builds</h2', '</ul>')
        trigger = "Unknown"
        upstream_urls = []
        if 'Started by upstream project' in body:
            upstream = extract_token(body, 'Started by upstream project', '</td')
            upstream_urls = re.findall(r'href="([^"]+[0-9]/)"', upstream)
            trigger = None
        if 'Started by GitHub push' in body:
            try:
                trigger = "commit_" + extract_token(body, 'commit: ', '<')
            except TypeError:
                trigger = "commit_unknown"
        if 'Started by user' in body:
            trigger = "started_by_user"
        downstream_urls = []
        if downstream_builds:
            downstream_urls = re.findall(r'href="([^"]+[0-9]/)"', downstream_builds)
        build = Build(url, host, name, build_number, start, duration, status, downstream_urls,
                      self.server_url, trigger, upstream_urls)
        return build

    def stats(self):
        """Compute statistics about the build tree."""
        start = stop = None
        duration = 0
        for build in self.builds.itervalues():
            if not build.start_t or not build.stop_t:
                continue
            if start is None or build.start_t < start:
                start = build.start_t
            if stop is None or build.stop_t > stop:
                stop = build.stop_t
            duration += build.duration_s
        elapsed = stop - start
        throughput = 0
        if elapsed.seconds:
            throughput = duration * 100. / elapsed.seconds
        ret = {'start': start,
               'stop': stop,
               'elapsed': elapsed,
               'duration': duration,
               'throughput': throughput,
               'count': len(self.builds)}
        return ret

    def clean_orphan(self, parent):
        """Remove build with no direct upstream."""
        for child in parent.children:
            if len(set(child.get_upstream()).intersection(set(self.builds.keys()))) == 0:
                parent.children.remove(child)
                if child.url in self.builds:
                    print "Removing orphean: %s" % child.url
                    self.builds.pop(child.url)
            self.clean_orphan(child)

    def list_builds(self, parent, builds):
        """Walk the tree to list the builds."""
        if parent.url not in builds:
            builds.append(parent.url)
        for build in parent.children:
            self.list_builds(build, builds)

    def clean(self, parent):
        if not self.options.explore:
            for i in range(10):
                # wtf is that :)
                self.clean_orphan(parent)
        all_builds = []
        self.list_builds(parent, all_builds)
        for url in self.builds.keys():
            if url not in all_builds:
                self.builds.pop(url)



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.test
   ~~~~~~~~~~~~~~~~

   Test cases
"""
import os
from tempfile import mkdtemp
from unittest import TestCase
from jenkviz.util import duration_to_second
from jenkviz.main import main


class JenkVizTestCase(TestCase):
    def test_duration(self):
        self.assertEquals(duration_to_second('1 min'), 60)
        self.assertEquals(duration_to_second('1 min 1 sec'), 61)
        self.assertEquals(duration_to_second('10 min 10 sec'), 610)
        self.assertEquals(duration_to_second('1 sec'), 1)
        self.assertEquals(duration_to_second('55 sec'), 55)
        self.assertEquals(duration_to_second('1 h 1 sec'), 3601)
        self.assertEquals(duration_to_second('1 h 1 min 1 sec'), 3661)

    def test_graphviz(self):
        tmp_dir = mkdtemp('_jenkviz')
        svg_file = os.path.join(tmp_dir, 'test.svg')
        ret = main(['jenkviz', 'crawl', 'http://foo/jenkins/job/nuxeo-features-master/72/', '--from-file=test-data', '-o', svg_file])
        self.assertEquals(ret, 0)

    def test_graphviz_direct(self):
        tmp_dir = mkdtemp('_jenkviz')
        svg_file = os.path.join(tmp_dir, 'test-direct.svg')
        ret = main(['jenkviz', 'crawl', 'http://foo/jenkins/job/nuxeo-features-master/72/', '--from-file=test-data', '-o', svg_file, '--direct'])
        self.assertEquals(ret, 0)



#!/usr/bin/env python
# -*- coding: utf_8 -*
__author__ = "Benoit Delbosc"
__copyright__ = "Copyright (C) 2012 Nuxeo SA <http://nuxeo.com>"
"""
   jenkviz.util
   ~~~~~~~~~~~~~~~~

   Misc utils
"""
import hashlib
import re
import logging
import pkg_resources
from datetime import datetime


def get_version():
    """Retrun the package version."""
    return pkg_resources.get_distribution('jenkviz').version


def init_logging(options):
    if hasattr(logging, '_bb_init'):
        return
    level = logging.INFO
    if options.verbose:
        level = logging.DEBUG
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M:%S',
                        filename=options.logfile,
                        filemode='w')
    console = logging.StreamHandler()
    console.setLevel(level)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)
    logging._bb_init = True
    print "Logging to " + options.logfile


def md5sum(filename):
    f = open(filename)
    md5 = hashlib.md5()
    while True:
        data = f.read(8192)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()


def str2id(filename):
    return re.sub(r"[^a-zA-Z0-9_]", r"_", filename)


# credits goes to Subways and Django folks
class BaseFilter(object):
    """Base filter."""
    def __ror__(self, other):
        return other  # pass-thru

    def __call__(self, other):
        return other | self


class truncate(BaseFilter):
    """Middle truncate string up to length."""
    def __init__(self, length=40, extra='...'):
        self.length = length
        self.extra = extra

    def __ror__(self, other):
        if len(other) > self.length:
            mid_size = (self.length - 3) / 2
            other = other[:mid_size] + self.extra + other[-mid_size:]
        return other


def duration_to_second(duration):
    """Convert jenkins duration into second"""
    if not duration:
        return None
    match = re.match('^(([0-9])+ h)? ?(([0-9]+) min)? ?(([0-9]+) sec)?$', duration)
    ret = 0
    if match and len(match.groups()) == 6:
        if match.group(2):
            ret += 3600 * int(match.group(2))
        if match.group(4):
            ret += 60 * int(match.group(4))
        if match.group(6):
            ret += int(match.group(6))
    return ret


def time_to_datetime(str_time):
    if str_time:
        return datetime.strptime(str_time, '%b %d, %Y %I:%M:%S %p')
    return None


def extract_token(text, tag_start, tag_end):
    """Extract a token from text, using the first occurence of
    tag_start and ending with tag_end. Return None if tags are not
    found."""
    start = text.find(tag_start)
    end = text.find(tag_end, start + len(tag_start))
    if start < 0 or end < 0:
        return None
    return text[start + len(tag_start):end]
